from typing import Annotated

from sqla_filter.base import BaseSorter
from sqla_filter.ordering import OrderingEnum, OrderingField
from sqla_filter.relationship import RelationshipInfo
from sqla_filter.unset import UNSET, Unset

from .models import Author, Book, User


class BookSorter(BaseSorter):
    created_at: Annotated[OrderingEnum | Unset, OrderingField(Book.created_at)] = UNSET
    created_at_nulls_last: Annotated[
        OrderingEnum | Unset,
        OrderingField(
            Book.created_at,
            nulls="last",
        ),
    ] = UNSET
    created_at_nulls_first: Annotated[
        OrderingEnum | Unset,
        OrderingField(
            Book.created_at,
            nulls="first",
        ),
    ] = UNSET
    author_user_first_name: Annotated[
        OrderingEnum | Unset,
        OrderingField(
            User.first_name,
            relationships=[
                RelationshipInfo(field=Book.authors),
                RelationshipInfo(field=Author.user),
            ],
        ),
    ] = UNSET
    author_alias: Annotated[
        OrderingEnum | Unset,
        OrderingField(Author.alias, relationship=RelationshipInfo(field=Book.authors)),
    ] = UNSET
    author_user_last_name: Annotated[
        OrderingEnum | Unset,
        OrderingField(
            User.last_name,
            relationships=[
                RelationshipInfo(field=Book.authors),
                RelationshipInfo(field=Author.user),
            ],
        ),
    ] = UNSET
    author_user_is_deleted: Annotated[
        OrderingEnum | Unset,
        OrderingField(
            User.is_deleted,
            relationships=[
                RelationshipInfo(field=Book.authors),
                RelationshipInfo(field=Author.user),
            ],
        ),
    ] = UNSET



import dataclasses
from collections.abc import Sequence
from typing import Any, Protocol

from sqlalchemy import ColumnElement
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.sql.operators import OperatorType

from .relationship import RelationshipInfo


class OperatorProtocol(Protocol):  # pragma: no cover
    def __call__(
        self,
        field: InstrumentedAttribute[Any],
        value: Any,  # noqa: ANN401
    ) -> ColumnElement[bool]: ...


@dataclasses.dataclass(slots=True)
class FilterField:
    _name: str = dataclasses.field(init=False, repr=False, hash=True)

    field: InstrumentedAttribute[Any]
    _: dataclasses.KW_ONLY
    operator: OperatorType | OperatorProtocol
    relationship: RelationshipInfo | None = None
    """For filter by relationship field"""

    relationships: Sequence[RelationshipInfo] | None = None
    """For filter by nested relationship field"""

    @property
    def name(self) -> str:  # pragma: no cover
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value



import pytest

from sqla_filter.unset import UNSET, Unset, or_unset


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (888, 888),
        ("888", "888"),
        (None, UNSET),
    ],
)
def test_or_unset(value: int | str | None, expected: int | str | Unset) -> None:
    assert or_unset(value) == expected



import dataclasses
from collections.abc import Iterable, Mapping
from typing import (
    Any,
    ClassVar,
    Literal,
    TypeVar,
    dataclass_transform,
    get_args,
    get_type_hints,
)

from sqlalchemy import Select, UnaryExpression
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.sql.functions import coalesce

from .filter_ import FilterField, RelationshipInfo
from .ordering import OrderingEnum, OrderingField
from .unset import Unset

_SelectClause = TypeVar("_SelectClause", bound=tuple[Any, ...])


@dataclass_transform(kw_only_default=True)
class BaseFilter:
    __sqla_filter_fields__: ClassVar[Mapping[str, FilterField]]

    def __init_subclass__(cls) -> None:
        _init_subclass(cls)

    def apply(self, stmt: Select[_SelectClause]) -> Select[_SelectClause]:
        for field_name, filter_ in self.__sqla_filter_fields__.items():
            value = getattr(self, field_name)

            if value is Unset.v:
                continue

            if filter_.relationship:
                stmt = _apply_join(stmt, relationship=filter_.relationship)

            for relationship in filter_.relationships or ():
                stmt = _apply_join(stmt, relationship=relationship)

            stmt = stmt.where(
                filter_.operator(
                    filter_.field,
                    value,
                ),  # pyright:ignore[reportArgumentType]
            )

        return stmt


@dataclass_transform(kw_only_default=True)
class BaseSorter:
    __sqla_filter_fields__: ClassVar[Mapping[str, OrderingField]]

    def __init_subclass__(cls) -> None:
        _init_subclass(cls)

    def apply(
        self,
        stmt: Select[_SelectClause],
        fields_priority: Iterable[OrderingField] | None = None,
    ) -> Select[_SelectClause]:
        all_fields = self.__sqla_filter_fields__
        if fields_priority:
            sorted_fields = {
                field.name: all_fields[field.name]
                for field in fields_priority
                if field.name in all_fields
            }
            remaining_fields = {
                field.name: all_fields[field.name]
                for field in all_fields.values()
                if field not in fields_priority
            }
            sorted_fields.update(remaining_fields)
            all_fields = sorted_fields

        for field_name, sorter in all_fields.items():
            value = getattr(self, field_name)

            if value is Unset.v:
                continue

            if sorter.relationship:
                stmt = _apply_join(stmt, relationship=sorter.relationship)

            for relationship in sorter.relationships or ():
                stmt = _apply_join(stmt, relationship=relationship)

            expr = _get_ordering_method(
                sorter.field,
                ordering=value,
                nulls=sorter.nulls,
            )
            stmt = stmt.order_by(expr)

        return stmt


def _init_subclass(cls: type[Any]) -> None:
    wrapped_cls = dataclasses.dataclass(cls)
    class_type_hints = get_type_hints(wrapped_cls, include_extras=True)
    del class_type_hints["__sqla_filter_fields__"]

    filter_fields = {}

    for field_name, raw_type in class_type_hints.items():
        args = get_args(raw_type)
        if not args:  # pragma: no cover
            continue

        sqla_filter_field = _get_sqla_filter_field(args)
        if sqla_filter_field is None:
            continue

        sqla_filter_field.name = field_name
        filter_fields[field_name] = sqla_filter_field
        setattr(wrapped_cls, field_name, sqla_filter_field)

    wrapped_cls.__sqla_filter_fields__ = filter_fields


def _get_sqla_filter_field(
    annotations: tuple[Any, ...],
) -> FilterField | OrderingField | None:
    main, *other = annotations

    for annotation in other:
        if isinstance(annotation, FilterField):
            return annotation

        if isinstance(annotation, OrderingField):
            if set(get_args(main)) != {OrderingEnum, Unset}:
                msg = "Annotate ordering field as Annotated[OrderingEnum | Unset]"
                raise TypeError(msg)

            return annotation

    return None


def _apply_join(
    stmt: Select[_SelectClause],
    *,
    relationship: RelationshipInfo,
) -> Select[_SelectClause]:
    return stmt.join(
        relationship.field,
        relationship.onclause,
        isouter=relationship.isouter,
        full=relationship.full,
    )


def _get_ordering_method(
    model_field: InstrumentedAttribute[Any] | coalesce[Any],
    *,
    ordering: OrderingEnum,
    nulls: Literal["first", "last"] | None = None,
) -> UnaryExpression[Any]:
    expr = model_field.asc() if ordering is OrderingEnum.asc else model_field.desc()

    match nulls:
        case "first":
            expr = expr.nulls_first()
        case "last":
            expr = expr.nulls_last()
        case _:
            pass

    return expr



from typing import Annotated, Any

from sqlalchemy import Select, select

from examples.models import Author, Book, User
from sqla_filter import (
    UNSET,
    BaseSorter,
    OrderingEnum,
    OrderingField,
    RelationshipInfo,
    Unset,
)
from sqla_filter.ordering import build_priorities


class BookSorter(BaseSorter):
    created_at: Annotated[OrderingEnum | Unset, OrderingField(Book.created_at)] = UNSET
    author_user_first_name: Annotated[
        OrderingEnum | Unset,
        OrderingField(
            User.first_name,
            relationships=[
                RelationshipInfo(field=Book.authors),
                RelationshipInfo(field=Author.user),
            ],
        ),
    ] = UNSET
    author_user_last_name: Annotated[
        OrderingEnum | Unset,
        OrderingField(
            User.last_name,
            relationships=[
                RelationshipInfo(field=Book.authors),
                RelationshipInfo(field=Author.user),
            ],
        ),
    ] = UNSET
    author_user_is_deleted: Annotated[
        OrderingEnum | Unset,
        OrderingField(
            User.is_deleted,
            relationships=[
                RelationshipInfo(field=Book.authors),
                RelationshipInfo(field=Author.user),
            ],
        ),
    ] = UNSET


def main() -> None:
    stmt = select(Book)
    print_stmt(stmt)

    sorter = BookSorter(
        author_user_last_name=OrderingEnum.desc,
        created_at=OrderingEnum.desc,
        author_user_is_deleted=OrderingEnum.desc,
    )

    stmt = sorter.apply(
        stmt,
        fields_priority=build_priorities(
            BookSorter.author_user_is_deleted,
            BookSorter.author_user_last_name,
        ),
    )
    print_stmt(stmt)


def print_stmt(stmt: Select[tuple[Any, ...]]) -> None:
    from sqlalchemy.dialects import postgresql

    print(  # noqa: T201
        stmt.compile(
            dialect=postgresql.dialect(),  # type: ignore[no-untyped-call]
            compile_kwargs={
                "literal_binds": True,
            },
        ),
        "\n",
    )


if __name__ == "__main__":
    main()



from collections.abc import Sequence
from dataclasses import dataclass
from datetime import datetime
from typing import Annotated
from uuid import UUID

from sqlalchemy.sql.operators import eq, ge, icontains_op, in_op, le

from sqla_filter import UNSET, BaseFilter, FilterField, RelationshipInfo, Unset

from .models import Author, Book, Review, User


@dataclass(frozen=True, slots=True)
class DateTimeInterval:
    from_: datetime
    to: datetime


class BookFilter(BaseFilter):
    ident: Annotated[UUID | Unset, FilterField(Book.id, operator=eq)] = UNSET

    created_at_from: Annotated[
        datetime | Unset,
        FilterField(Book.created_at, operator=ge),
    ] = UNSET
    created_at_to: Annotated[
        datetime | Unset,
        FilterField(Book.created_at, operator=le),
    ] = UNSET

    created_at_between: Annotated[
        DateTimeInterval | Unset,
        FilterField(
            Book.created_at,
            operator=lambda field, value: field.between(value.from_, value.to),
        ),
    ] = UNSET

    author_ids: Annotated[
        Sequence[UUID] | Unset,
        FilterField(
            Author.id,
            operator=in_op,
            relationship=RelationshipInfo(field=Book.authors),
        ),
    ] = UNSET
    author_user_id: Annotated[
        UUID | Unset,
        FilterField(
            User.id,
            operator=eq,
            relationships=[
                RelationshipInfo(field=Book.authors),
                RelationshipInfo(field=Author.user),
            ],
        ),
    ] = UNSET
    review_ids: Annotated[
        Sequence[UUID] | Unset,
        FilterField(
            Review.id,
            operator=in_op,
            relationship=RelationshipInfo(field=Book.reviews),
        ),
    ] = UNSET
    review_content_contains: Annotated[
        str | Unset,
        FilterField(
            Review.content,
            operator=icontains_op,
            relationship=RelationshipInfo(field=Book.reviews),
        ),
    ] = UNSET
    process_manually_field: str | Unset = UNSET
